#
#  Copyright 2010-2016 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import logging as log
import os
import re
import stat
import sys
from argparse import ArgumentParser

from six.moves.configparser import ConfigParser
from six.moves.urllib.parse import urlsplit

from Pegasus.tools.worker_utils import logger

try:
    import boto3
    import botocore
except ImportError as e:
    sys.stderr.write("ERROR: Unable to load boto3 library: %s\n" % e)
    exit(1)

# do not use http proxies for S3
if "http_proxy" in os.environ:
    del os.environ["http_proxy"]

# Don't let apple hijack our cacerts
os.environ["OPENSSL_X509_TEA_DISABLE"] = "1"

COMMANDS = {
    "ls": "List the contents of a bucket",
    "mkdir": "Create a bucket in S3",
    "rm": "Delete a key from S3",
    "put": "Upload a key to S3 from a file",
    "get": "Download a key from S3 to a file",
    "cp": "Copy keys remotely",
    "help": "Print this message",
}


KB = 1024
MB = 1024 * KB
GB = 1024 * MB
TB = 1024 * GB


def configure_logging(verbose, debug):

    if verbose or debug:
        if verbose:
            level = log.INFO

        if debug:
            level = log.DEBUG

        log.basicConfig(
            level=level,
            format="%(asctime)s:%(levelname)s:%(name)s(%(lineno)d): %(message)s",
        )


def human_size(size):
    if size >= TB:
        return "{0:6.1f}TB".format(size / float(TB))
    elif size >= GB:
        return "{0:6.1f}GB".format(size / float(GB))
    elif size >= MB:
        return "{0:6.1f}MB".format(size / float(MB))
    elif size >= KB:
        return "{0:6.1f}KB".format(size / float(KB))
    else:
        return "{0:6.0f}B".format(size)


# see https://docs.aws.amazon.com/general/latest/gr/s3.html
LOCATIONS = {
    "s3.amazonaws.com": "",
    "s3.us-east-1.amazonaws.com": "us-east-1",
    "s3.us-east-2.amazonaws.com": "us-east-2",
    "s3-us-west-1.amazonaws.com": "us-west-1",
    "s3-us-west-2.amazonaws.com": "us-west-2",
    "s3.af-south-1.amazonaws.com": "af-south-1",
    "s3-ca-central-1.amazonaws.com": "ca-central-1",
    "s3.ca-west-1.amazonaws.com": "ca-west-1",
    "s3-eu-west-1.amazonaws.com": "EU",
    "s3.eu-central-1.amazonaws.com": "eu-central-1",
    "s3.eu-west-1.amazonaws.com": "eu-west-1",
    "s3.eu-west-2.amazonaws.com": "eu-west-2",
    "s3.eu-west-3.amazonaws.com": "eu-west-3",
    "s3.eu-south-1.amazonaws.com": "eu-south-1",
    "s3.eu-south-2.amazonaws.com": "eu-south-2",
    "s3.eu-north-1.amazonaws.com": "eu-north-1",
    "s3.eu-central-2.amazonaws.com": "eu-central-2",
    "s3.il-central-1.amazonaws.com": "il-central-1",
    "s3.me-south-1.amazonaws.com": "me-south-1",
    "s3.me-central-1.amazonaws.com": "me-central-1",
    "s3.sa-east-1.amazonaws.com": "sa-east-1",
    "s3.us-gov-east-1.amazonaws.com": "us-gov-east-1",
    "s3.us-gov-west-1.amazonaws.com": "us-gov-west-1",
    "s3.ap-east-1.amazonaws.com": "ap-east-1",
    "s3.ap-south-2.amazonaws.com": "ap-south-2",
    "s3-ap-southeast-1.amazonaws.com": "ap-southeast-1",
    "s3-ap-southeast-2.amazonaws.com": "ap-southeast-2",
    "s3.ap-southeast-3.amazonaws.com": "ap-southeast-3",
    "s3.ap-southeast-4.amazonaws.com": "ap-southeast-4",
    "s3.ap-southeast-5.amazonaws.com": "ap-southeast-5",
    "s3-ap-northeast-1.amazonaws.com": "ap-northeast-1",
    "s3-ap-northeast-2.amazonaws.com": "ap-northeast-2",
    "s3-ap-northeast-3.amazonaws.com": "ap-northeast-3",
    "s3-ap-south-1.amazonaws.com": "ap-south-1",
    "s3-sa-east-1.amazonaws.com": "sa-east-1",
}

DEFAULT_CONFIG = {
    "batch_delete": str(True),
    "batch_delete_size": str(1000),
}


def fix_file(url):
    if url.startswith("file://"):
        url = url.replace("file:", "")
    return url


def has_wildcards(string):
    if string is None:
        return False
    wildcards = "*?[]"
    for c in wildcards:
        if c in string:
            return True
    return False


def get_path_for_key(bucket, searchkey, key, output):
    # We have to strip any trailing / off the keys so that they can match
    # Also, if a key is None, then convert it to an empty string
    key = "" if key is None else key.rstrip("/")
    searchkey = "" if searchkey is None else searchkey.rstrip("/")

    # If output ends with a /, then we need to add a name onto it
    if output.endswith("/"):
        name = bucket if searchkey == "" else os.path.basename(searchkey)
        output = os.path.join(output, name)

    if searchkey == key:
        # If they are the same, then return the new output path
        return output
    else:
        # Otherwise we need to compute the relative path and add it
        relpath = os.path.relpath(key, searchkey)
        return os.path.join(output, relpath)


DEFAULT_CREDENTIAL_PATH = "~/.pegasus/credentials.conf"
OLD_DEFAULT_CREDENTIAL_PATH = "~/.pegasus/s3cfg"


def get_config(options):
    PEGASUS_CREDENTIALS = os.getenv("PEGASUS_CREDENTIALS", None)
    S3CFG = os.getenv("S3CFG", None)
    if options.config:
        # Command-line overrides everything
        cfg = options.config
    elif PEGASUS_CREDENTIALS is not None:
        # Environment variable overrides defaults
        cfg = PEGASUS_CREDENTIALS
    elif S3CFG is not None:
        # Environment variable overrides defaults
        cfg = S3CFG
    else:
        # New default
        new_default = os.path.expanduser(DEFAULT_CREDENTIAL_PATH)
        if os.path.isfile(new_default):
            cfg = new_default
        else:
            # If the new default doesn't exist, try the old default
            cfg = os.path.expanduser(OLD_DEFAULT_CREDENTIAL_PATH)

    if not os.path.isfile(cfg):
        raise Exception("Config file not found")

    log.info("Found config file: %s" % cfg)

    # Make sure nobody else can read the file
    mode = os.stat(cfg).st_mode
    if mode & (stat.S_IRWXG | stat.S_IRWXO):
        raise Exception("Permissions of config file %s are too liberal" % cfg)

    config = ConfigParser(DEFAULT_CONFIG)
    config.read(cfg)

    return config


def parse_endpoint(uri):
    result = urlsplit(uri)

    kwargs = {
        "is_secure": result.scheme == "https",
        "host": result.hostname,
        "port": result.port,
        "path": result.path,
    }

    location = LOCATIONS.get(result.hostname, "")

    return kwargs, location


def get_s3_client(config, uri):
    if not config.has_section(uri.site):
        raise Exception("Config file has no section for site '%s'" % uri.site)

    if not config.has_section(uri.ident):
        raise Exception("Config file has no section for identity '%s'" % uri.ident)

    endpoint = config.get(uri.site, "endpoint")
    region_name = config.get(uri.site, "region", fallback=None)
    aws_access_key_id = config.get(uri.ident, "access_key")
    aws_secret_access_key = config.get(uri.ident, "secret_key")

    # PM-1978 make sure endpoint in the config file is
    # consistent with region for AWS endpoints only
    if region_name is not None and "amazonaws.com" in endpoint:
        kwargs, endpoint_region = parse_endpoint(endpoint)
        if endpoint_region != region_name:
            # KV: maybe should be an error?
            logger.warning(
                "Inconsistency between AWS endpoint {} and region {} specified in the config file".format(
                    endpoint, region_name
                )
            )

    # what about s3s????

    return (
        boto3.client(
            "s3",
            endpoint_url=endpoint,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
        ),
        region_name,
    )


def get_create_bucket_configuration(region_name):
    config = {}
    # don't set any create bucket configuration for us-east-1
    # https://github.com/aws/aws-sdk-js/issues/3647
    if region_name is not None and region_name != "us-east-1":
        config["CreateBucketConfiguration"] = {"LocationConstraint": region_name}
    return config


def is_bucket_available(s3_client, bucket):
    is_available = False
    try:
        s3_client.head_bucket(Bucket=bucket)
        is_available = True
    except botocore.exceptions.ClientError as e:
        code = e.response["Error"]["Code"]

        # Not Found
        if code == "404":
            is_available = True

        # Anything other than 403 Forbidden, won't handle
        elif code != "403":
            raise e

    return is_available


def read_command_file(path):
    tokenizer = re.compile(r"\s+")
    f = open(path, "r")
    try:
        for line in f:
            line = line.strip()
            if len(line) == 0:
                continue
            if line.startswith("#"):
                continue
            yield tokenizer.split(line)
    finally:
        f.close()


class S3URI:
    def __init__(self, user, site, bucket=None, key=None, secure=False):
        self.user = user
        self.site = site
        self.ident = "%s@%s" % (user, site)
        self.bucket = bucket
        self.key = key
        self.secure = secure

    def __repr__(self):
        if self.secure:
            uri = "s3s://%s" % self.ident
        else:
            uri = "s3://%s" % self.ident
        if self.bucket is not None:
            uri += "/%s" % self.bucket
        if self.key is not None:
            uri += "/%s" % self.key
        return uri


def parse_uri(uri):
    "Parse S3 uri into an S3URI object"

    # The only valid schemes are s3s:// and s3://
    if uri.startswith("s3s://"):
        secure = True
    elif uri.startswith("s3://"):
        secure = False
    else:
        raise Exception("Invalid URL scheme: %s" % (uri))

    # Need to do this because urlparse does not recognize
    # custom URI schemes. Replace our scheme with http.
    # The actual scheme used isn't important as long as
    # urlsplit recognizes it.
    if secure:
        http = uri.replace("s3s://", "http://")
    else:
        http = uri.replace("s3://", "http://")
    result = urlsplit(http)

    # The user should not be specifying a query part unless
    # they are trying to use the ? wildcard. If they do use
    # the ? wildcard, then urlsplit thinks it is the query
    # separator. In that case, we put the path and query
    # back together.
    if "?" in uri:
        path = "?".join([result.path, result.query]).strip()
    else:
        path = result.path.strip()

    # The path should be empty, /BUCKET or /BUCKET/KEY
    if path.startswith("/"):
        path = path[1:]
    if len(path) == 0:
        bucket = None
        key = None
    else:
        comp = path.split("/", 1)
        bucket = comp[0]
        if len(comp) == 1:
            key = None
        elif comp[1] == "":
            key = None
        else:
            key = comp[1]

    # We require the username part
    user = result.username
    if user is None:
        raise Exception("User missing from URL: %s" % uri)

    if result.port is None:
        site = result.hostname
    else:
        site = "%s:%s" % (result.hostname, result.port)

    return S3URI(user, site, bucket, key, secure)


def ls(args):
    config = get_config(args)
    uri = parse_uri(args.url)

    s3, region_name = get_s3_client(config, uri)

    if uri.bucket:
        # list keys in bucket
        try:
            keys = s3.list_objects_v2(
                Bucket=uri.bucket, Prefix=uri.key if uri.key else "", FetchOwner=True
            )
        except s3.exceptions.NoSuchBucket:
            raise Exception("Invalid bucket: {}".format(uri.bucket))
        except botocore.exceptions.ClientError as e:
            # endpoint may also raise this for invalid bucket name
            if e.response["Error"]["Code"] == "InvalidBucketName":
                raise Exception("Invalid bucket: {}".format(uri.bucket))
            else:
                raise e

        if keys.get("Contents"):
            for content in keys["Contents"]:
                key = content["Key"]

                if args.long_format:
                    size = (
                        human_size(content["Size"])
                        if args.human_sized
                        else "{0:13d}".format(content["Size"])
                    )
                    last_modified = content["LastModified"]
                    owner = content["Owner"]["DisplayName"]
                    storage_class = content["StorageClass"]
                    print(
                        "\t{owner:15s} {size} {modified} {storage_class:24s} {name}".format(
                            owner=owner,
                            size=size,
                            modified=last_modified,
                            storage_class=storage_class,
                            name=key,
                        )
                    )
                else:
                    print("\t{}".format(key))
    else:
        # list buckets
        buckets = s3.list_buckets()
        for b in buckets["Buckets"]:
            print("\t{}".format(b["Name"]))


def cp(args):
    config = get_config(args)

    srcs = [parse_uri(uri) for uri in args.srcs]
    dest = parse_uri(args.dest)

    # If there is more than one source, then the destination must be
    # a bucket and not a bucket+key.
    if len(srcs) > 1 and dest.key is not None:
        raise Exception("Destination must be a bucket if there are multiple sources")

    # Validate all the URI pairs
    for src in srcs:
        # The source URI must have a key
        if src.key is None:
            raise Exception("Source URL does not contain a key: %s" % src)

        # Each source must have the same identity as the destination.
        # Copying from one account to another, or one region to another,
        # is not allowed.
        if src.ident != dest.ident:
            raise Exception(
                "Identities for source and destination "
                "do not match: %s -> %s" % (src, dest)
            )

    # using the first source (from the checks above, it is garuanteed that
    # identities all match)
    s3, region_name = get_s3_client(config, srcs[0])

    # Create the bucket if the user requested it and it does not exist
    if args.create:
        can_create = True
        try:
            s3.head_bucket(Bucket=dest.bucket)

            # no exception, we already own this bucket
            can_create = False
        except botocore.exceptions.ClientError as e:
            code = e.response["Error"]["Code"]

            # 403 forbidden means bucket already taken
            if code == "403":
                raise Exception(
                    "Bucket: {} is already taken. Unable to create bucket.".format(
                        dest.bucket
                    )
                )

            # 404 not found means bucket can be created
            elif code != "404":
                raise e

        if can_create:
            s3.create_bucket(
                Bucket=dest.bucket, **get_create_bucket_configuration(region_name),
            )

    # ensure that none of the keys in srcs exist in dest
    if not args.force:
        if dest.key == None:
            for src in srcs:
                try:
                    s3.head_object(Bucket=dest.bucket, Key=src.key)
                    raise Exception(
                        "Key: {key} already exists in destination bucket: {bucket}, (see --force)".format(
                            key=src.key, bucket=dest.bucket
                        )
                    )
                except botocore.exceptions.ClientError as e:
                    error_code = e.response["ResponseMetadata"]["HTTPStatusCode"]

                    if error_code != 404:
                        raise e

        else:
            assert len(srcs) == 1
            src = srcs[0]
            try:
                s3.head_object(Bucket=dest.bucket, Key=dest.key)
                raise Exception(
                    "Key: {key} already exists in destination bucket: {bucket}, (see --force)".format(
                        key=dest.key, bucket=dest.bucket
                    )
                )
            except botocore.exceptions.ClientError as e:
                error_code = e.response["ResponseMetadata"]["HTTPStatusCode"]
                if error_code != 404:
                    raise e

    if dest.key == None:
        for src in srcs:
            s3.copy(
                CopySource={"Bucket": src.bucket, "Key": src.key},
                Bucket=dest.bucket,
                Key=src.key,
            )
    else:
        assert len(srcs) == 1
        src = srcs[0]
        s3.copy(
            CopySource={"Bucket": src.bucket, "Key": src.key},
            Bucket=dest.bucket,
            Key=dest.key,
        )


def mkdir(args):
    uri = parse_uri(args.url)

    if uri.bucket is None:
        raise Exception("URL for mkdir must contain a bucket: %s" % arg)
    if uri.key is not None:
        raise Exception("URL for mkdir cannot contain a key: %s" % arg)

    config = get_config(args)
    s3, region_name = get_s3_client(config, uri)
    can_create = True
    try:
        s3.head_bucket(Bucket=uri.bucket)

        # no exception, we already own this bucket
        can_create = False
    except botocore.exceptions.ClientError as e:
        code = e.response["Error"]["Code"]

        # 403 forbidden means bucket already taken
        if code == "403":
            raise Exception(
                "Bucket: {} is already taken. Unable to create bucket.".format(
                    uri.bucket
                )
            )

        # 404 not found means bucket can be created
        elif code != "404":
            raise e

    if can_create:
        s3.create_bucket(
            Bucket=uri.bucket, **get_create_bucket_configuration(region_name),
        )
    else:
        log.warning(
            "Bucket: {} exists and is already owned by user: {}".format(
                uri.bucket, uri.ident
            )
        )


def rm(args):
    if args.url is None and args.file is None:
        print("Specify URL")
        sys.exit(1)

    uris = []
    if args.file:
        for rec in read_command_file(args.file):
            if len(rec) != 1:
                raise Exception("Invalid record: %s" % rec)
            uris.append(rec[0])
    else:
        uris.append(args.url)

    buckets = {}
    for uri in uris:
        uri = parse_uri(uri)
        if uri.bucket is None:
            raise Exception("URL for rm must contain a bucket: %s" % uri)
        if uri.key is None:
            raise Exception("URL for rm must contain a key: %s" % uri)

        bid = "%s/%s" % (uri.ident, uri.bucket)
        buri = S3URI(uri.user, uri.site, uri.bucket, uri.secure)

        if bid not in buckets:
            buckets[bid] = (buri, set())
        buckets[bid][1].add(uri)

    config = get_config(args)

    for bucket in buckets:

        # Connect to the bucket
        log.info("Deleting keys from bucket %s" % bucket)
        uri, keys = buckets[bucket]
        try:
            s3, region_name = get_s3_client(config, uri)

            keys_to_delete = {k.key for k in keys}

            # check that all keys to be deleted exist if force is not set
            if not args.force:
                bucket_contents = {
                    obj["Key"]
                    for obj in s3.list_objects_v2(Bucket=uri.bucket)["Contents"]
                }
                if not keys_to_delete.issubset(bucket_contents):
                    raise Exception(
                        "Some keys to delete do not exist in bucket: {}; use --force to ignore non existing keys".format(
                            uri.bucket
                        )
                    )

            log.info("Deleting %d keys" % len(keys_to_delete))

            batch_delete = config.getboolean(uri.site, "batch_delete")

            if batch_delete:
                log.info("Using batch deletes")

                # Delete the keys in batches
                batch_delete_size = config.getint(uri.site, "batch_delete_size")
                log.info("batch_delete_size: %d" % batch_delete_size)
                batch = []
                for k in keys_to_delete:
                    batch.append(k)
                    if len(batch) == batch_delete_size:
                        log.info("Deleting batch of %d keys" % len(batch))

                        resp = s3.delete_objects(
                            Bucket=uri.bucket,
                            Delete={"Objects": [{"Key": item} for item in batch]},
                        )

                        if not len(resp["Deleted"]) == len(batch):
                            raise Exception(
                                "Incomplete batch delete, some keys were not successfully deleted."
                            )

                        batch = []

                # Delete the final batch
                if len(batch) > 0:
                    log.info("Deleting batch of %d keys" % len(batch))
                    resp = s3.delete_objects(
                        Bucket=uri.bucket,
                        Delete={"Objects": [{"Key": item} for item in batch]},
                    )

                    if not len(resp["Deleted"]) == len(batch):
                        raise Exception(
                            "Incomplete batch delete, some keys were not successfully deleted."
                        )

            else:
                for key_name in keys_to_delete:
                    log.info("Deleting %s" % key_name)
                    s3.delete_object(Bucket=uri.bucket, Key=key_name)

        except s3.exceptions.NoSuchBucket:
            raise Exception("Invalid bucket: {}".format(uri.bucket))
        except botocore.exceptions.ClientError as e:
            # endpoint may also raise this for invalid bucket name
            if e.response["Error"]["Code"] == "InvalidBucketName":
                log.error("Invalid bucket: {}".format(uri.bucket))

            raise e


def get_key_for_path(path, infile, outkey):
    if outkey is None or outkey == "":
        raise Exception("invalid key: '%s'" % outkey)

    if not path.startswith("/"):
        raise Exception("path '%s' should be absolute")

    path = path.rstrip("/")
    infile = infile.rstrip("/")

    if not infile.startswith(path):
        raise Exception("file '%s' is not relative to '%s'" % (infile, path))

    if outkey.endswith("/"):
        name = os.path.basename(path)
        outkey = outkey + name

    relpath = os.path.relpath(infile, path)
    if relpath != ".":
        return os.path.join(outkey, relpath)
    else:
        return outkey


def put(args):
    path = fix_file(args.file)
    url = args.url

    if not os.path.exists(path):
        raise Exception("No such file or directory: {}".format(path))

    if os.path.isdir(path):
        raise Exception("FILE: %s is a directory. FILE must be a file." % path)

    log.info("Attempting to upload {}".format(path))

    # Validate URL
    uri = parse_uri(url)
    if uri.bucket is None:
        raise Exception("URL for put must have a bucket: %s" % url)
    if uri.key is None:
        uri.key = os.path.basename(path)

    config = get_config(args)

    # get s3 client with associated endpoint
    s3, region_name = get_s3_client(config, uri)

    # Create the bucket if the user requested it and it does not exist
    if args.create_bucket:
        can_create = True
        try:
            s3.head_bucket(Bucket=uri.bucket)

            # no exception, we already own this bucket
            can_create = False
        except botocore.exceptions.ClientError as e:
            code = e.response["Error"]["Code"]

            if code != "404":
                log.error(code)
                raise e

        if can_create:
            s3.create_bucket(
                Bucket=uri.bucket, **get_create_bucket_configuration(region_name),
            )

    if not args.force:
        # check if all keys do not yet exist
        # accepted method of checking for existence of a key
        key_already_exists = False

        try:
            s3.head_object(Bucket=uri.bucket, Key=uri.key)
            key_already_exists = True
        except s3.exceptions.ClientError as e:
            code = e.response["ResponseMetadata"]["HTTPStatusCode"]

            if code != 404:
                log.error(code)
                raise e

        if key_already_exists:
            raise Exception(
                "Key: {} already exists. Try --force to overwrite".format(path)
            )

    try:
        key = path if uri.key is None else uri.key
        s3.upload_file(path, uri.bucket, key)
        log.info(
            "Uploaded file: {file} to bucket: {bucket} as key: {key}".format(
                file=path, bucket=uri.bucket, key=key
            )
        )
    except boto3.exceptions.S3UploadFailedError:
        raise Exception(
            "Failed to upload file: {file} to bucket: {bucket} as key: {key}".format(
                file=path, bucket=uri.bucket, key=key
            )
        )


def get(args):
    uri = parse_uri(args.url)

    if uri.bucket is None:
        raise Exception("URL must contain a bucket: %s" % args[0])
    if uri.key is None:
        raise Exception("URL must contain a key")

    if args.file:
        output = fix_file(args.file)
    else:
        output = os.path.basename(uri.key.rstrip("/"))

    log.info("Downloading %s" % uri)

    config = get_config(args)
    s3, region_name = get_s3_client(config, uri)

    try:
        s3.download_file(Bucket=uri.bucket, Key=uri.key, Filename=output)
    except s3.exceptions.NoSuchBucket:
        raise Exception("Invalid bucket: {}".format(uri.bucket))
    except botocore.exceptions.ClientError as e:
        # endpoint may also raise this for invalid bucket name
        log.error("Response error code: {}".format(e.response["Error"]["Code"]))
        raise e

    log.info("Download: {} complete".format(uri))


# --- Handle Command Line Arguments --------------------------------------------
def parse_args(args):
    parser = ArgumentParser(prog="pegasus-s3")

    # add top level arguments
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Enable verbose mode"
    )

    parser.add_argument("-d", "--debug", action="store_true", help="Enable debug mode")

    parser.add_argument(
        "-C", "--conf", dest="config", default=None, help="Path to configuration file"
    )

    # create a subparser that will handle s3 commands
    subparser = parser.add_subparsers(dest="cmd", help="Available s3 commands")

    # create subcommands for each of the supported s3 commands
    # LS command --------------------------
    parser_ls = subparser.add_parser("ls")
    parser_ls.add_argument("url", metavar="URL", help="URL to be looked up")
    parser_ls.add_argument(
        "-l",
        "--long",
        dest="long_format",
        action="store_true",
        default=False,
        help="Use long listing format",
    )
    parser_ls.add_argument(
        "-H",
        "--human-sized",
        dest="human_sized",
        action="store_true",
        default=False,
        help="Use human readable sizes",
    )
    parser_ls.set_defaults(func=ls)

    # MKDIR command -----------------------------
    parser_mkdir = subparser.add_parser("mkdir")
    parser_mkdir.add_argument(
        "url", metavar="URL", help="URL specifying bucket to be created"
    )
    parser_mkdir.set_defaults(func=mkdir)

    # RM command --------------------------
    parser_rm = subparser.add_parser("rm")
    parser_rm.add_argument(
        "url",
        metavar="URL",
        nargs="?",
        default=None,
        help="URL specifying key to be removed",
    )

    parser_rm.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=False,
        help="Ignore nonexistent keys",
    )
    parser_rm.add_argument(
        "-F",
        "--file",
        dest="file",
        action="store",
        default=None,
        help="File containing a list of URLs to delete",
    )
    parser_rm.set_defaults(func=rm)

    # PUT command ---------------------------
    parser_put = subparser.add_parser("put")
    parser_put.add_argument("file", metavar="FILE", help="The file to be uploaded")

    parser_put.add_argument(
        "url", metavar="URL", help="URL to which the file will be uploaded"
    )

    parser_put.add_argument(
        "-b",
        "--create-bucket",
        dest="create_bucket",
        action="store_true",
        default=False,
        help="Create the destination bucket if it does not already exist",
    )

    parser_put.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=False,
        help="Overwrite key if it already exists",
    )
    parser_put.set_defaults(func=put)

    # GET command --------------------------
    parser_get = subparser.add_parser("get")
    parser_get.add_argument("url", metavar="URL", help="URL of the key to download")

    parser_get.add_argument(
        "file",
        nargs="?",
        default=None,
        metavar="FILE",
        help="File that key will be downloaded as",
    )
    parser_get.set_defaults(func=get)

    # CP command -------------------------
    parser_cp = subparser.add_parser("cp")
    parser_cp.add_argument(
        "srcs", nargs="+", metavar="SRC", help="Sources to copy from"
    )

    parser_cp.add_argument("dest", metavar="DST", help="Destination to copy to")

    parser_cp.add_argument(
        "-c",
        "--create-dest",
        dest="create",
        action="store_true",
        default=False,
        help="Create destination bucket if it does not exist",
    )
    parser_cp.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=False,
        help="If DEST key exists, then overwrite it",
    )
    parser_cp.set_defaults(func=cp)

    return parser, parser.parse_args(args)


# --- Entrypoint ---------------------------------------------------------------
def main():
    parser, args = parse_args(sys.argv[1:])

    configure_logging(args.verbose, args.debug)

    if args.cmd is None:
        parser.print_usage()
    else:
        try:
            args.func(args)
        except Exception as e:
            if args.verbose or args.debug:
                log.exception("Encountered error:")
            else:
                print(e)

            sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Pegasus utility for checking file integrity after transfers

Usage: pegasus-integrity-check [options]
"""
from __future__ import print_function

import glob
import json
import logging
import optparse
import os
import pprint
import sys
import time
from datetime import datetime

from Pegasus.tools import worker_utils as utils

##
#  Copyright 2007-2017 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the 'License');
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an 'AS IS' BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##


# for some reason, sometimes grp is missing, but we can do without it
try:
    import grp
    import pwd
except Exception:
    pass


# see https://www.python.org/dev/peps/pep-0469/
try:
    dict.iteritems
except AttributeError:
    # Python 3
    def itervalues(d):
        return iter(d.values())

    def iteritems(d):
        return iter(d.items())


else:
    # Python 2
    def itervalues(d):
        return d.itervalues()

    def iteritems(d):
        return d.iteritems()


__author__ = "Mats Rynge <rynge@isi.edu>"


# --- classes -----------------------------------------------------------------


class Singleton(type):
    """Implementation of the singleton pattern"""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
            cls.lock = threading.Lock()
        return cls._instances[cls]


# --- global variables ----------------------------------------------------------------

prog_dir = os.path.realpath(os.path.join(os.path.dirname(sys.argv[0])))
prog_base = os.path.split(sys.argv[0])[1]  # Name of this program

logger = logging.getLogger("PegasusIntegrity")

multipart_fh = None

count_succeeded = 0
count_failed = 0
total_timing = 0.0

# --- functions ----------------------------------------------------------------


def setup_logger(debug_flag):

    # log to the console
    console = logging.StreamHandler()

    # default log level - make logger/console match
    logger.setLevel(logging.INFO)
    console.setLevel(logging.INFO)

    # debug - from command line
    if debug_flag:
        logger.setLevel(logging.DEBUG)
        console.setLevel(logging.DEBUG)

    # formatter
    formatter = logging.Formatter("Integrity check: %(message)s")
    console.setFormatter(formatter)
    logger.addHandler(console)
    logger.debug("Logger has been configured")


def backticks(cmd_line):
    """
    what would a python program be without some perl love?
    """
    return (
        subprocess.Popen(cmd_line, shell=True, stdout=subprocess.PIPE)
        .communicate()[0]
        .decode("utf-8")
    )


def json_object_decoder(obj):
    """
    utility function used by json.load() to parse some known objects into equivalent Python objects
    """
    if "type" in obj and obj["type"] == "transfer":
        t = Transfer()
        # src
        for surl in obj["src_urls"]:
            priority = None
            if "priority" in surl:
                priority = int(surl["priority"])
            t.add_src(surl["site_label"], surl["url"], priority)
        for durl in obj["dest_urls"]:
            priority = None
            if "priority" in durl:
                priority = int(durl["priority"])
            t.add_dst(durl["site_label"], durl["url"], priority)
        return t
    elif "type" in obj and obj["type"] == "mkdir":
        m = Mkdir()
        m.set_url(obj["target"]["site_label"], obj["target"]["url"])
        return m
    elif "type" in obj and obj["type"] == "remove":
        r = Remove()
        r.set_url(obj["target"]["site_label"], obj["target"]["url"])
        if "recursive" in obj["target"]:
            r.set_recursive(obj["target"]["recursive"])
        return r
    return obj


def read_meta_data(f):
    """
    Reads transfers from the new JSON based input format
    """
    data = []
    try:
        fp = open(f, "r")
        data = json.load(fp)
        fp.close()
    except Exception as err:
        logger.critical("Error parsing the meta data: " + str(err))
    return data


def myexit(rc):
    """
    system exit without a stack trace
    """
    try:
        sys.exit(rc)
    except SystemExit:
        sys.exit(rc)


def generate_sha256(fname):
    """
    Generates a sha256 hash for the given file
    """

    tools = utils.Tools()
    tools.find("openssl", "version", "([0-9]+\.[0-9]+\.[0-9]+)")
    tools.find("sha256sum", "--version", "([0-9]+\.[0-9]+)")

    if tools.full_path("openssl"):
        cmd = tools.full_path("openssl")
        if tools.version_comparable("openssl") < "001001000":
            cmd += " sha -sha256"
        else:
            cmd += " sha256"
        cmd += " " + fname
    elif tools.full_path("sha256sum"):
        cmd = tools.full_path("sha256sum")
        cmd += " " + fname + " | sed 's/ .*//'"
    else:
        logger.error("openssl or sha256sum not found!")
        return None

    if not os.path.exists(fname):
        logger.error("File " + fname + " does not exist")
        return None

    # generate the checksum
    tc = utils.TimedCommand(cmd, log_cmd=False, log_outerr=False)
    tc.run()
    if tc.get_exit_code() != 0:
        logger.error("Unable to determine sha256: " + tc.get_outerr())
        return None

    sha256 = tc.get_outerr()
    sha256 = sha256.strip()
    sha256 = sha256[-64:]
    if len(sha256) != 64:
        logger.warn("Unable to determine sha256 of " + fname)

    return sha256


def iso8601(ts):
    """
    Formats a UNIX timestamp in ISO 8601 format
    """
    dt = datetime.utcfromtimestamp(ts)
    return dt.isoformat()


def generate_yaml(lfn, pfn):
    """
    Generates kickstart yaml for the given file
    """

    ts_start = time.time()
    sha256 = generate_sha256(pfn)
    if sha256 is None:
        return None
    ts_end = time.time()

    return "      sha256: %s\n      checksum_timing: %.3f\n" % (
        sha256,
        ts_end - ts_start,
    )


def generate_fullstat_yaml(lfn, pfn):
    """
    Generates kickstart yaml for the given file
    """

    ts_start = time.time()
    sha256 = generate_sha256(pfn)
    if sha256 is None:
        return None
    ts_end = time.time()

    yaml = ""
    yaml += '    "%s":\n' % (lfn)
    yaml += "      comment: entry generated by pegasus-integrity\n"
    yaml += '      file_name: "%s"\n' % (pfn)

    try:
        s = os.stat(pfn)
    except Exception:
        # if we can't stat, just return
        return ""

    uname = ""
    try:
        uname = pwd.getpwuid(s.st_uid).pw_name
    except Exception:
        pass
    gname = ""
    try:
        gname = grp.getgrgid(s.st_gid).gr_name
    except Exception:
        pass
    yaml += (
        "      mode: 0o%o\n"
        "      size: %d\n"
        "      inode: %d\n"
        "      nlink: %d\n"
        "      mtime: %s\n"
        "      atime: %s\n"
        "      ctime: %s\n"
        "      uid: %d\n"
        "      user: %s\n"
        "      gid: %d\n"
        "      group: %s\n"
        "      output: True\n"
        % (
            s.st_mode,
            s.st_size,
            s.st_ino,
            s.st_nlink,
            iso8601(s.st_mtime),
            iso8601(s.st_atime),
            iso8601(s.st_ctime),
            s.st_uid,
            uname,
            s.st_gid,
            gname,
        )
    )
    yaml += "      sha256: %s\n" "      checksum_timing: %.3f\n" % (
        sha256,
        ts_end - ts_start,
    )
    return yaml


def check_integrity(fname, lfn, meta_data, print_timings):
    """
    Checks the integrity of a file given a set of metadata
    """

    global count_succeeded
    global count_failed
    global total_timing

    ts_start = time.time()

    if lfn is None or lfn == "":
        lfn = fname

    # find the expected checksum in the metadata
    expected_sha256 = None
    for entry in meta_data:
        if entry["_id"] == lfn:
            if "_attributes" in entry and "checksum.value" in entry["_attributes"]:
                expected_sha256 = entry["_attributes"]["checksum.value"]
    if expected_sha256 is None:
        logger.error("No checksum in the meta data for " + lfn)
        return False

    current_sha256 = generate_sha256(fname)

    ts_end = time.time()

    total_timing += ts_end - ts_start

    if print_timings:
        check_info_yaml(
            lfn,
            fname,
            current_sha256,
            expected_sha256,
            current_sha256 == expected_sha256,
        )

    # now compare them
    if current_sha256 != expected_sha256:
        logger.error(
            "%s: Expected checksum (%s) does not match the calculated checksum (%s) (timing: %.3f)"
            % (fname, expected_sha256, current_sha256, ts_end - ts_start)
        )
        count_failed += 1
        return False

    count_succeeded += 1
    return True


def multipart_out(s):
    """
    returns a multipart fh if required
    """
    global multipart_fh

    if "PEGASUS_MULTIPART_DIR" not in os.environ:
        return

    if multipart_fh == -1:
        # previous error
        return
    elif multipart_fh is None:
        try:
            multipart_fh = open(
                "%s/%d-integrity"
                % (os.environ["PEGASUS_MULTIPART_DIR"], int(time.time())),
                "w",
            )
        except Exception as e:
            logger.error("Unable to write stats to $PEGASUS_MULTIPART_DIR: " + str(e))
            multipart_fh = -1

    # ready to write
    try:
        multipart_fh.write(s)
    except Exception:
        pass


def check_info_yaml(lfn, pfn, sha256, expected_sha256, success):
    """
    prints stats for monitord
    """
    multipart_out(
        ('  - lfn: "%s"\n' '    pfn: "%s"\n' "    sha256: %s\n" "    success: %s\n")
        % (lfn, pfn, sha256, str(success))
    )
    if not success:
        multipart_out("    sha256_expected: %s\n" % expected_sha256)


def dump_summary_yaml():
    """
    outputs a timing block for Pegasus monitoring
    """
    multipart_out(
        (
            "- integrity_summary:\n"
            "    succeeded: %d\n"
            "    failed: %d\n"
            "    duration: %.3f\n"
        )
        % (count_succeeded, count_failed, total_timing)
    )


# --- main ----------------------------------------------------------------------------


def main():
    global threads
    global stats_start
    global stats_end
    global symlink_file_transfer

    # dup stderr onto stdout
    sys.stderr = sys.stdout

    # Configure command line option parser
    prog_usage = "usage: %s [options]" % (prog_base)
    parser = optparse.OptionParser(usage=prog_usage)

    parser.add_option(
        "",
        "--generate",
        action="store",
        dest="generate_files",
        help="Generate a SHA256 hash for a set of files",
    )
    parser.add_option(
        "",
        "--generate-yaml",
        action="store",
        dest="generate_yaml",
        help="Generate hashes for the given file, output to kickstart yaml.",
    )
    parser.add_option(
        "",
        "--generate-fullstat-yaml",
        action="store",
        dest="generate_fullstat_yaml",
        help="Generate hashes for the given file, output to kickstart yaml.",
    )
    parser.add_option(
        "",
        "--generate-xmls",
        action="store",
        dest="generate_xmls",
        help="Generate hashes for the given file, output to kickstart xml.",
    )
    parser.add_option(
        "",
        "--generate-fullstat-xmls",
        action="store",
        dest="generate_fullstat_xmls",
        help="Generate hashes for the given file, output to kickstart xml.",
    )
    parser.add_option(
        "",
        "--verify",
        action="store",
        dest="verify_files",
        help="Verify the hash for the given file.",
    )
    parser.add_option(
        "",
        "--print-timings",
        action="store_true",
        dest="print_timings",
        help="Display timing data after verifying files",
    )
    parser.add_option(
        "",
        "--debug",
        action="store_true",
        dest="debug",
        help="Enables debugging output.",
    )

    # Parse command line options
    (options, args) = parser.parse_args()
    setup_logger(options.debug)

    # sanity checks
    if (
        sum(
            [
                options.generate_files is not None,
                options.generate_yaml is not None,
                options.generate_fullstat_yaml is not None,
                options.generate_xmls is not None,
                options.generate_fullstat_xmls is not None,
                options.verify_files is not None,
            ]
        )
        != 1
    ):
        logger.error(
            "One, and only one, of --generate-* and --verify needs to be specified"
        )
        parser.print_help()
        sys.exit(1)

    if options.generate_files:
        for f in str.split(options.generate_files, ":"):
            results = generate_sha256(f)
            if not results:
                myexit(1)
            print(results + "  " + f)

    elif options.generate_yaml:
        for f in str.split(options.generate_yaml, ":"):
            # lfn can be encoded in the file name in the format lfn=pfn
            lfn = None
            pfn = f
            if "=" in f:
                (lfn, pfn) = str.split(f, "=", 1)
            results = generate_yaml(lfn, pfn)
            if not results:
                myexit(1)
            print(results)

    elif options.generate_fullstat_yaml:
        for f in str.split(options.generate_fullstat_yaml, ":"):
            # lfn can be encoded in the file name in the format lfn=pfn
            lfn = None
            pfn = f
            if "=" in f:
                (lfn, pfn) = str.split(f, "=", 1)
            results = generate_fullstat_yaml(lfn, pfn)
            if not results:
                myexit(1)
            # if lfn is not None and 'KICKSTART_INTEGRITY_DATA' in os.environ:
            if "KICKSTART_INTEGRITY_DATA" in os.environ:
                f = open(os.environ["KICKSTART_INTEGRITY_DATA"], "a")
                f.write(results)
                f.close()
            else:
                print(results)

    elif options.verify_files:

        if options.print_timings:
            multipart_out("- integrity_verification_attempts:\n")

        # read all the .meta files in the current working dir
        meta_data = []
        for meta_file in glob.glob("*.meta"):
            logger.debug("Loading metadata from %s" % (meta_file))
            all_md = read_meta_data(meta_file)
            for entry in all_md:
                meta_data.append(entry)

        if options.debug:
            pprint.PrettyPrinter(indent=4).pprint(meta_data)

        # the files can be provided directly, or via stdin
        files = ""
        if options.verify_files == "stdin":
            files = sys.stdin.read().strip()
        else:
            files = options.verify_files

        # now check the files
        exit_code = 0
        for f in str.split(files, ":"):
            # lfn can be encoded in the file name in the format lfn=pfn
            lfn = None
            pfn = f
            if "=" in f:
                (lfn, pfn) = str.split(f, "=", 1)
            results = check_integrity(pfn, lfn, meta_data, options.print_timings)
            if not results:
                exit_code = 1

        if options.print_timings:
            dump_summary_yaml()

        myexit(exit_code)

    myexit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
This is a tool used by pegasus-transfer to do transfers between
Globus Online endpoints
"""

#
#  Copyright 2016 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import json
import logging
import optparse
import os
import re
import signal
import sys
from datetime import datetime, timedelta

import globus_sdk

# --- global variables ----------------------------------------------------------------

prog_dir = os.path.realpath(os.path.join(os.path.dirname(sys.argv[0])))
prog_base = os.path.split(sys.argv[0])[1]  # Name of this program

logger = logging.getLogger("Pegasus")

client = None
transfer_client = None
task_id = None

# --- functions ----------------------------------------------------------------


def setup_logger(debug_flag):

    # log to the console
    console = logging.StreamHandler()

    # default log level - make logger/console match
    logger.setLevel(logging.INFO)
    console.setLevel(logging.INFO)

    # debug - from command line
    if debug_flag:
        logger.setLevel(logging.DEBUG)
        console.setLevel(logging.DEBUG)

    # formatter
    formatter = logging.Formatter("%(asctime)s %(levelname)7s:  %(message)s")
    console.setFormatter(formatter)
    logger.addHandler(console)
    logger.debug("Logger has been configured")


def prog_sigint_handler(signum, frame):
    logger.warn("Exiting due to signal %d" % (signum))
    if transfer_client is not None:
        cancel_task(transfer_client, task_id)
    sys.exit(1)


def acquire_clients(request):
    # connect to the service
    client = globus_sdk.NativeAppAuthClient(request["client_id"])

    if request["transfer_rt"] is None:
        authorizer = globus_sdk.AccessTokenAuthorizer(request["transfer_at"])
    else:
        authorizer = globus_sdk.RefreshTokenAuthorizer(
            request["transfer_rt"],
            client,
            access_token=request["transfer_at"],
            expires_at=int(request["transfer_at_exp"]),
        )

    transfer_client = globus_sdk.TransferClient(
        authorizer=authorizer, transport_params={"max_retries": 0}
    )
    return client, transfer_client


def activate_endpoint(transfer_client, endpoint):
    """
    Try to Auto-Activate a Globus Online endpoint - only auto-activate is allowed for now
    """
    logger.info("Activating " + endpoint)
    try:
        res = transfer_client.endpoint_autoactivate(endpoint, if_expires_in=3600)
        if res["code"] == "AutoActivationFailed":
            logger.critical('Endpoint "%s" requires manual activation' % endpoint)
            logger.critical(
                "Please log into Globus Online and activate the endpoint there"
            )
            sys.exit(1)
    except globus_sdk.TransferAPIError as e:
        logger.critical('Endpoint "%s" auto-activation ERROR' % endpoint)
        logger.critical(str(e))
        sys.exit(1)


def wait_for_task(transfer_client, task_id, acceptable_faults=None):
    """
    Wait for a task to complete
    """
    logger.info("Waiting for transfer to complete")
    wait_timeout = 60
    loop_counter = 0

    #### SHOULD CHECK FOR ERRORS ####
    while not transfer_client.task_wait(task_id, timeout=wait_timeout):
        loop_counter += 1
        logger.info(
            "Globus transfer task %s is still running (%d seconds)"
            % (task_id, loop_counter * wait_timeout)
        )
        task_errors = transfer_client.task_event_list(
            task_id=task_id, limit=20, query_params={"filter": "is_error:1"}
        )
        for error in task_errors:
            details = re.sub(r"\n|\r", " ", error["description"])
            if re.search(r"System error in mkdir.*File exists", details):
                logger.info("Ignoring mkdir error: " + details)
            else:
                raise RuntimeError(
                    "Error on globus transfer task %s at %s: %s"
                    % (task_id, error["time"], details)
                )

    return None


def cancel_task(transfer_client, task_id):
    """
    Cancel a task - useful when a transfer has errors or we catch a signal
    """
    logger.info("Canceling transfer")
    try:
        transfer_client.cancel_task(task_id)
        logger.info("Globus transfer task %s has been canceled" % task_id)
    except Exception:
        pass


def mkdir(request):
    """
    operation_mkdir doesn't support recursive creation of directories
    """
    # global so that we can use it in signal handlers
    global client
    global transfer_client
    global task_id

    # connect to the service
    client, transfer_client = acquire_clients(request)

    # make sure we can auto-activate the endpoints
    activate_endpoint(transfer_client, request["endpoint"])

    for f in request["files"]:
        base_path = ""
        child_dirs = []

        # find base_path with operation_ls
        base_path = f
        found = False
        while (not found) and base_path != "/":
            found = True
            base_path, dir_name = os.path.split(base_path)
            if dir_name not in ["", "/"]:
                child_dirs.append(dir_name)
            try:
                response = transfer_client.operation_ls(
                    request["endpoint"], path=base_path, query_params={"limit": 2}
                )
            except globus_sdk.TransferAPIError as e:
                logger.warn("Finding existing parent dir for mkdir " + f)
                logger.warn(str(e))
                found = False

        child_dirs.reverse()
        path = base_path
        for child in child_dirs:
            if path[-1] == "/":
                path += child
            else:
                path = path + "/" + child

            try:
                transfer_client.operation_mkdir(request["endpoint"], path)
            except globus_sdk.TransferAPIError as e:
                if e.code == "ExternalError.MkdirFailed.Exists":
                    logger.warn("Directory already exists: " + path)
                else:
                    raise RuntimeError(str(e))
    logger.info("Mkdir complete")


def transfer(request):
    """
    takes a transfer specification parsed from json:
    {
      "client_id":       "globus oauth client id",
      "src_endpoint":    "rynge#obelix",
      "dst_endpoint":    "rynge#workflow",
      "transfer_at":     "secret_auth_token",
      "transfer_rt":     "secret_refresh_auth_token",
      "transfer_at_exp": "secret_auth_token_expiration",
      "recursive":       [True | False],
      "files":[
         {"src":"/etc/hosts","dst":"/tmp/foobar.txt"},
         {"src":"/etc/hosts","dst":"/tmp/foobar-2.txt"}
        ],
    }
    """

    # global so that we can use it in signal handlers
    global client
    global transfer_client
    global task_id

    # connect to the service
    client, transfer_client = acquire_clients(request)

    # make sure we can auto-activate the endpoints
    activate_endpoint(transfer_client, request["src_endpoint"])
    activate_endpoint(transfer_client, request["dst_endpoint"])

    label = None
    if "PEGASUS_WF_UUID" in os.environ and "PEGASUS_DAG_JOB_ID" in os.environ:
        label = os.environ["PEGASUS_WF_UUID"] + " - " + os.environ["PEGASUS_DAG_JOB_ID"]

    # set up a new data transfer
    deadline = datetime.utcnow() + timedelta(hours=24)
    transfer_data = globus_sdk.TransferData(
        transfer_client,
        source_endpoint=request["src_endpoint"],
        destination_endpoint=request["dst_endpoint"],
        label=label,
        deadline=deadline,
        notify_on_succeeded=False,
        notify_on_failed=False,
        notify_on_inactive=False,
    )

    for pair in request["files"]:
        transfer_data.add_item(pair["src"], pair["dst"])

    # finalize and submit the transfer
    transfer_result = transfer_client.submit_transfer(transfer_data)
    task_id = transfer_result["task_id"]

    # how many faults will we accept before giving up?
    # acceptable_faults = min(100, len(request["files"]) * 3)

    # wait for the task to complete, and see the tasks and
    # endpoint ls change
    try:
        wait_for_task(transfer_client, task_id)
    except Exception as err:
        logger.error(err)
        cancel_task(transfer_client, task_id)
        sys.exit(1)
    logger.info("Transfer complete")


def remove(request):
    """
    removes files on a remote Globus Online endpoint - API is not complete, so transfer
    0 byte files instead of actually deleting anything
    """

    # global so that we can use it in signal handlers
    global client
    global transfer_client
    global task_id

    # connect to the service
    client, transfer_client = acquire_clients(request)

    activate_endpoint(transfer_client, request["endpoint"])

    label = None
    if "PEGASUS_WF_UUID" in os.environ and "PEGASUS_DAG_JOB_ID" in os.environ:
        label = os.environ["PEGASUS_WF_UUID"] + " - " + os.environ["PEGASUS_DAG_JOB_ID"]

    # set up a new delete transfer
    deadline = datetime.utcnow() + timedelta(hours=24)
    del_data = globus_sdk.DeleteData(
        transfer_client,
        endpoint=request["endpoint"],
        label=label,
        recursive=request["recursive"],
        deadline=deadline,
        notify_on_succeeded=False,
        notify_on_failed=False,
        notify_on_inactive=False,
    )

    for f in request["files"]:
        del_data.add_item(f)

    # finalize and submit the transfer
    delete_result = transfer_client.submit_delete(del_data)
    task_id = delete_result["task_id"]

    # how many faults will we accept before giving up?
    # acceptable_faults = min(100, len(request["files"]) * 3)

    # wait for the task to complete, and see the tasks and
    # endpoint ls change
    try:
        wait_for_task(transfer_client, task_id)
    except Exception as err:
        logger.error(err)
        cancel_task(transfer_client, task_id)
        sys.exit(1)
    logger.info("Delete complete")


def main():

    # Configure command line option parser
    prog_usage = "usage: %s [options]" % (prog_base)
    parser = optparse.OptionParser(usage=prog_usage)

    parser.add_option(
        "--mkdir", action="store_true", dest="mkdir", help="Select mkdir mode"
    )
    parser.add_option(
        "--transfer", action="store_true", dest="transfer", help="Select transfer mode"
    )
    parser.add_option(
        "--remove", action="store_true", dest="remove", help="Select remove mode"
    )
    parser.add_option(
        "--file",
        action="store",
        dest="file",
        help="File containing GO URL pairs to be transferred",
    )
    parser.add_option(
        "-d",
        "--debug",
        action="store_true",
        dest="debug",
        help="Enables debugging output",
    )

    # Parse command line options
    (options, args) = parser.parse_args()
    setup_logger(options.debug)

    if not options.file:
        logger.critical("An input file has to be given with --file")
        sys.exit(1)

    # get the json
    with open(options.file) as data_file:
        data = json.load(data_file)

    # Die nicely when asked to (Ctrl+C, system shutdown)
    signal.signal(signal.SIGINT, prog_sigint_handler)
    signal.signal(signal.SIGTERM, prog_sigint_handler)

    logger.info("Globus SDK Version: %s" % globus_sdk.__version__)

    if options.mkdir:
        mkdir(data)
    elif options.transfer:
        transfer(data)
    elif options.remove:
        remove(data)
    else:
        logger.critical("Please specify one of: --mkdir, --transfer, --remove")
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from __future__ import print_function

import os
from argparse import ArgumentParser

import globus_sdk
from globus_sdk.scopes import (
    GCSCollectionScopeBuilder,
    GCSEndpointScopeBuilder,
    TransferScopes,
)

try:
    from configparser import ConfigParser
except Exception:
    from ConfigParser import ConfigParser

client_id = "d7382f5a-4ea3-4b69-b094-99c392fc820d"
config_file = os.path.expanduser("~/.pegasus/globus.conf")

parser = ArgumentParser(
    description="Initialize Globus OAuth Tokens - SDK Version {}".format(
        globus_sdk.__version__
    )
)
parser.add_argument(
    "-p", "--permanent", action="store_true", help="request a refreshable token",
)
parser.add_argument(
    "-e",
    "--endpoints",
    nargs="*",
    default=[],
    help="list of endpoint uuids to acquire manage_collections consent",
)
parser.add_argument(
    "-c",
    "--collections",
    nargs="*",
    default=[],
    help="list of collection uuids to acquire data_access consent",
)
parser.add_argument(
    "-d",
    "--domains",
    nargs="*",
    default=[],
    help="list of domain requirements which must be satisfied by the identities under the globus user account",
)


args = parser.parse_args()

config = ConfigParser()
config.add_section("oauth")
config.set("oauth", "client_id", client_id)

pegasus_scopes = [TransferScopes.all]
for c in args.collections:
    sb = GCSCollectionScopeBuilder(c, known_scopes=[TransferScopes.all])
    pegasus_scopes.append("{0}[*{1}]".format(TransferScopes.all, sb.data_access))
    pegasus_scopes.append(sb.data_access)

for e in args.endpoints:
    sb = GCSEndpointScopeBuilder(e)
    pegasus_scopes.append(sb.manage_collections)

required_domains = None
for d in args.domains:
    if not required_domains:
        required_domains = [d]
    else:
        required_domains.append(d)

client = globus_sdk.NativeAppAuthClient(client_id)
if args.permanent:
    client.oauth2_start_flow(refresh_tokens=True, requested_scopes=pegasus_scopes)
else:
    client.oauth2_start_flow(requested_scopes=pegasus_scopes)

authorize_url = client.oauth2_get_authorize_url(
    session_required_single_domain=required_domains, prompt="login"
)
print("Please go to this URL and login: {}".format(authorize_url))

get_input = getattr(__builtins__, "raw_input", input)
auth_code = get_input("Please enter the code you get after login here: ").strip()
token_response = client.oauth2_exchange_code_for_tokens(auth_code)

transfer_data = token_response.by_resource_server["transfer.api.globus.org"]

# get tokens as strings and their expiration timestamp
transfer_at = str(transfer_data["access_token"])
transfer_rt = str(transfer_data["refresh_token"])
transfer_at_exp = str(transfer_data["expires_at_seconds"])

config.set("oauth", "transfer_at", transfer_at)  # transfer access token
config.set("oauth", "transfer_rt", transfer_rt)  # transfer refresh token
config.set("oauth", "transfer_at_exp", transfer_at_exp)  # transfer expiration time

cfg = open(config_file, "w+")
config.write(cfg)
cfg.close()



# -*- coding: utf-8 -*-

##
#  Copyright 2007-2011 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##

"""Provides common functions used by all workflow programs."""


from __future__ import print_function

import logging
import os
import re
import subprocess
import sys
import tempfile
import threading
import time

import six
from six.moves.builtins import int

# Module variables
logger = logging.getLogger("Pegasus")


def force_str(s):
    """
    force s to be str - should work both in py2 and py3
    """
    if six.PY3:
        return s.decode()
    return s


def backticks(cmd_line):
    """
    what would a python program be without some perl love?
    """
    return force_str(
        subprocess.Popen(cmd_line, shell=True, stdout=subprocess.PIPE).communicate()[0]
    )


class TimedCommand(object):
    """ Provides a shell callout with a timeout """

    def __init__(
        self,
        cmd,
        cmd_display=None,
        env_overrides={},
        timeout_secs=6 * 60 * 60,
        log_cmd=True,
        log_outerr=True,
        cwd=None,
    ):
        self._cmd = cmd
        if cmd_display is None:
            self._cmd_display = cmd
        else:
            self._cmd_display = cmd_display
        self._env_overrides = env_overrides.copy()
        self._timeout_secs = timeout_secs
        self._log_cmd = log_cmd
        self._log_outerr = log_outerr
        self._process = None
        self._out_file = None
        self._outerr = ""
        self._duration = 0.0
        self._cwd = cwd

    def run(self):
        def target():

            # custom environment for the subshell
            sub_env = os.environ.copy()
            for key, value in six.iteritems(self._env_overrides):
                logger.debug("ENV override: %s = %s" % (key, value))
                sub_env[key] = value

            self._process = subprocess.Popen(
                self._cmd,
                shell=True,
                env=sub_env,
                stdout=self._out_file,
                stderr=subprocess.STDOUT,
                preexec_fn=os.setpgrp,
                cwd=self._cwd,
            )
            self._process.communicate()

        if self._log_cmd or logger.isEnabledFor(logging.DEBUG):
            logger.info(self._cmd_display)

        sys.stdout.flush()

        # temp file for the stdout/stderr
        self._out_file = tempfile.TemporaryFile(prefix="pegasus-", suffix=".out")

        ts_start = time.time()

        thread = threading.Thread(target=target)
        thread.start()

        thread.join(self._timeout_secs)
        if thread.is_alive():
            # do our best to kill the whole process group
            try:
                # os.killpg did not work in all environments
                # os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)
                kill_cmd = "kill -TERM -%d" % (os.getpgid(self._process.pid))
                kp = subprocess.Popen(kill_cmd, shell=True)
                kp.communicate()
                self._process.terminate()
            except Exception:
                pass
            thread.join()
            # log the output
            self._out_file.seek(0)
            stdout = str.strip(force_str(self._out_file.read()))
            if len(stdout) > 0:
                logger.info(stdout)
            self._out_file.close()
            raise RuntimeError(
                "Command timed out after %d seconds: %s"
                % (self._timeout_secs, self._cmd_display)
            )

        self._duration = time.time() - ts_start

        # log the output
        self._out_file.seek(0)
        self._outerr = str.strip(force_str(self._out_file.read()))
        if self._log_outerr and len(self._outerr) > 0:
            logger.info(self._outerr)
        self._out_file.close()

        if self._process.returncode != 0:
            print(self.get_outerr())
            raise RuntimeError(
                "Command exited with non-zero exit code (%d): %s"
                % (self._process.returncode, self._cmd_display)
            )

    def get_outerr(self):
        """
        returns the combined stdout and stderr from the command
        """
        return self._outerr

    def get_exit_code(self):
        """
        returns the exit code from the process
        """
        return self._process.returncode

    def get_duration(self):
        """
        returns the timing of the command (seconds)
        """
        return self._duration


class Tools(object):
    """
    Singleton for detecting and maintaining tools we depend on
    """

    # singleton
    instance = None

    def __init__(self):
        if not Tools.instance:
            Tools.instance = Tools.__Tools()

    def __getattr__(self, name):
        return getattr(self.instance, name)

    class __Tools:

        _info = {}

        def __init__(self):
            self.lock = threading.Lock()

        def find(
            self, executable, version_arg=None, version_regex=None, path_prepend=None
        ):

            self.lock.acquire()
            try:
                if executable in self._info:
                    if self._info[executable] is None:
                        return None
                    return self._info[executable]["full_path"]

                logger.debug(
                    "Trying to detect availability/location of tool: %s" % (executable)
                )

                # initialize the global tool info for this executable
                self._info[executable] = {}
                self._info[executable]["full_path"] = None
                self._info[executable]["version"] = None
                self._info[executable]["version_major"] = None
                self._info[executable]["version_minor"] = None
                self._info[executable]["version_patch"] = None

                # figure out the full path to the executable
                path_entries = os.environ["PATH"].split(":")
                if "" in path_entries:
                    path_entries.remove("")
                # if we have PEGASUS_HOME set, and try to invoke a Pegasus tool, prepend
                if "PEGASUS_HOME" in os.environ and executable.find("pegasus-") == 0:
                    path_entries.insert(0, os.environ["PEGASUS_HOME"] + "/bin")
                if path_prepend is not None:
                    for entry in path_prepend:
                        path_entries.insert(0, entry)

                # now walk the path
                full_path = None
                for entry in path_entries:
                    full_path = entry + "/" + executable
                    if os.path.isfile(full_path) and os.access(full_path, os.X_OK):
                        break
                    full_path = None

                if full_path is None:
                    logger.info(
                        "Command '%s' not found in the current environment"
                        % (executable)
                    )
                    self._info[executable] = None
                    return self._info[executable]
                self._info[executable]["full_path"] = full_path

                # version
                if version_regex is None:
                    version = "N/A"
                else:
                    version = backticks(executable + " " + version_arg + " 2>&1")
                    version = version.replace("\n", "")
                    re_version = re.compile(version_regex)
                    result = re_version.search(version)
                    if result:
                        version = result.group(1)
                    self._info[executable]["version"] = version

                # if possible, break up version into major, minor, patch
                re_version = re.compile("([0-9]+)\.([0-9]+)(\.([0-9]+)){0,1}")
                result = re_version.search(version)
                if result:
                    self._info[executable]["version_major"] = int(result.group(1))
                    self._info[executable]["version_minor"] = int(result.group(2))
                    self._info[executable]["version_patch"] = result.group(4)
                if (
                    self._info[executable]["version_patch"] is None
                    or self._info[executable]["version_patch"] == ""
                ):
                    self._info[executable]["version_patch"] = None
                else:
                    self._info[executable]["version_patch"] = int(
                        self._info[executable]["version_patch"]
                    )

                logger.info(
                    "Tool found: %s   Version: %s   Path: %s"
                    % (executable, version, full_path)
                )
                return self._info[executable]["full_path"]
            finally:
                self.lock.release()

        def full_path(self, executable):
            """ Returns the full path to a given executable """
            self.lock.acquire()
            try:
                if executable in self._info and self._info[executable] is not None:
                    return self._info[executable]["full_path"]
                return None
            finally:
                self.lock.release()

        def major_version(self, executable):
            """ Returns the detected major version given executable """
            self.lock.acquire()
            try:
                if executable in self._info and self._info[executable] is not None:
                    return self._info[executable]["version_major"]
                return None
            finally:
                self.lock.release()

        def minor_version(self, executable):
            """ Returns the detected minor version given executable """
            self.lock.acquire()
            try:
                if executable in self._info and self._info[executable] is not None:
                    return self._info[executable]["version_minor"]
                return None
            finally:
                self.lock.release()

        def version_comparable(self, executable):
            """ Returns the detected comparable version given executable """
            self.lock.acquire()
            try:
                if executable in self._info and self._info[executable] is not None:
                    return "%03d%03d%03d" % (
                        int(self._info[executable]["version_major"]),
                        int(self._info[executable]["version_minor"]),
                        int(self._info[executable]["version_patch"]),
                    )
                return None
            finally:
                self.lock.release()



#!/usr/bin/env python3

import os
import subprocess
import sys

# only expand PYTHONPATH if it not already set
if "PEGASUS_PYTHONPATH_SET" not in os.environ:
    bin_dir = os.path.normpath(os.path.join(os.path.dirname(sys.argv[0])))
    pegasus_config = os.path.abspath(os.path.join(bin_dir, "pegasus-config"))
    lib_dir = subprocess.Popen(
        [pegasus_config, "--noeoln", "--python"], stdout=subprocess.PIPE, shell=False
    ).communicate()[0]
    lib_ext_dir = subprocess.Popen(
        [pegasus_config, "--noeoln", "--python-externals"],
        stdout=subprocess.PIPE,
        shell=False,
    ).communicate()[0]

    sys.path.insert(0, lib_ext_dir)
    sys.path.insert(0, lib_dir)


from Pegasus import s3  # isort:skip

s3.main()
